@@ -1,19 +1,16 @@
 use std::{
-    future::Future,
+    future::{poll_fn, Future},
     io,
     pin::Pin,
     task::{Context, Poll},
 };
 
 use bytes::Bytes;
-use proto::{Chunk, Chunks, ConnectionError, ReadableError, StreamId};
+use proto::{Chunk, Chunks, ClosedStream, ConnectionError, ReadableError, StreamId};
 use thiserror::Error;
 use tokio::io::ReadBuf;
 
-use crate::{
-    connection::{ConnectionRef, UnknownStream},
-    VarInt,
-};
+use crate::{connection::ConnectionRef, VarInt};
 
 /// A stream that can only be used to receive data
 ///
@@ -21,54 +18,36 @@ use crate::{
 /// - A variant of [`ReadError`] has been yielded by a read call
 /// - [`stop()`] was called explicitly
 ///
-/// # Closing a stream
+/// # Cancellation
 ///
-/// When a stream is expected to be closed gracefully the sender should call
-/// [`SendStream::finish`].  However there is no guarantee the connected [`RecvStream`] will
-/// receive the "finished" notification in the same QUIC frame as the last frame which
-/// carried data.
+/// A `read` method is said to be *cancel-safe* when dropping its future before the future becomes
+/// ready cannot lead to loss of stream data. This is true of methods which succeed immediately when
+/// any progress is made, and is not true of methods which might need to perform multiple reads
+/// internally before succeeding. Each `read` method documents whether it is cancel-safe.
 ///
-/// Even if the application layer logic already knows it read all the data because it does
-/// its own framing, it should still read until it reaches the end of the [`RecvStream`].
-/// Otherwise it risks inadvertently calling [`RecvStream::stop`] if it drops the stream.
-/// And calling [`RecvStream::stop`] could result in the connected [`SendStream::finish`]
-/// call failing with a [`WriteError::Stopped`] error.
+/// # Common issues
 ///
-/// For example if exactly 10 bytes are to be read, you still need to explicitly read the
-/// end of the stream:
+/// ## Data never received on a locally-opened stream
 ///
-/// ```no_run
-/// # use quinn::{SendStream, RecvStream};
-/// # async fn func(
-/// #     mut send_stream: SendStream,
-/// #     mut recv_stream: RecvStream,
-/// # ) -> anyhow::Result<()>
-/// # {
-/// // In the sending task
-/// send_stream.write(&b"0123456789"[..]).await?;
-/// send_stream.finish().await?;
+/// Peers are not notified of streams until they or a later-numbered stream are used to send
+/// data. If a bidirectional stream is locally opened but never used to send, then the peer may
+/// never see it. Application protocols should always arrange for the endpoint which will first
+/// transmit on a stream to be the endpoint responsible for opening it.
 ///
-/// // In the receiving task
-/// let mut buf = [0u8; 10];
-/// let data = recv_stream.read_exact(&mut buf).await?;
-/// if recv_stream.read_to_end(0).await.is_err() {
-///     // Discard unexpected data and notify the peer to stop sending it
-///     let _ = recv_stream.stop(0u8.into());
-/// }
-/// # Ok(())
-/// # }
-/// ```
+/// ## Data never received on a remotely-opened stream
 ///
-/// An alternative approach, used in HTTP/3, is to specify a particular error code used with `stop`
-/// that indicates graceful receiver-initiated stream shutdown, rather than a true error condition.
-///
-/// [`RecvStream::read_chunk`] could be used instead which does not take ownership and
-/// allows using an explit call to [`RecvStream::stop`] with a custom error code.
+/// Verify that the stream you are receiving is the same one that the server is sending on, e.g. by
+/// logging the [`id`] of each. Streams are always accepted in the same order as they are created,
+/// i.e. ascending order by [`StreamId`]. For example, even if a sender first transmits on
+/// bidirectional stream 1, the first stream yielded by [`Connection::accept_bi`] on the receiver
+/// will be bidirectional stream 0.
 ///
 /// [`ReadError`]: crate::ReadError
 /// [`stop()`]: RecvStream::stop
 /// [`SendStream::finish`]: crate::SendStream::finish
 /// [`WriteError::Stopped`]: crate::WriteError::Stopped
+/// [`id`]: RecvStream::id
+/// [`Connection::accept_bi`]: crate::Connection::accept_bi
 #[derive(Debug)]
 pub struct RecvStream {
     conn: ConnectionRef,
@@ -92,6 +71,8 @@ impl RecvStream {
     /// Read data contiguously from the stream.
     ///
     /// Yields the number of bytes read into `buf` on success, or `None` if the stream was finished.
+    ///
+    /// This operation is cancel-safe.
     pub async fn read(&mut self, buf: &mut [u8]) -> Result<Option<usize>, ReadError> {
         Read {
             stream: self,
@@ -102,7 +83,7 @@ impl RecvStream {
 
     /// Read an exact number of bytes contiguously from the stream.
     ///
-    /// See [`read()`] for details.
+    /// See [`read()`] for details. This operation is *not* cancel-safe.
     ///
     /// [`read()`]: RecvStream::read
     pub async fn read_exact(&mut self, buf: &mut [u8]) -> Result<(), ReadExactError> {
@@ -113,7 +94,25 @@ impl RecvStream {
         .await
     }
 
-    fn poll_read(
+    /// Attempts to read from the stream into buf.
+    ///
+    /// On success, returns Poll::Ready(Ok(num_bytes_read)) and places data in
+    /// the buf. If no data was read, it implies that EOF has been reached.
+    ///
+    /// If no data is available for reading, the method returns Poll::Pending
+    /// and arranges for the current task (via cx.waker()) to receive a notification
+    /// when the stream becomes readable or is closed.
+    pub fn poll_read(
+        &mut self,
+        cx: &mut Context,
+        buf: &mut [u8],
+    ) -> Poll<Result<usize, ReadError>> {
+        let mut buf = ReadBuf::new(buf);
+        ready!(self.poll_read_buf(cx, &mut buf))?;
+        Poll::Ready(Ok(buf.filled().len()))
+    }
+
+    fn poll_read_buf(
         &mut self,
         cx: &mut Context,
         buf: &mut ReadBuf<'_>,
@@ -153,6 +152,8 @@ impl RecvStream {
     ///
     /// Slightly more efficient than `read` due to not copying. Chunk boundaries do not correspond
     /// to peer writes, and hence cannot be used as framing.
+    ///
+    /// This operation is cancel-safe.
     pub async fn read_chunk(
         &mut self,
         max_length: usize,
@@ -166,7 +167,14 @@ impl RecvStream {
         .await
     }
 
-    /// Foundation of [`Self::read_chunk`]
+    /// Attempts to read a chunk from the stream.
+    ///
+    /// On success, returns `Poll::Ready(Ok(Some(chunk)))`. If `Poll::Ready(Ok(None))`
+    /// is returned, it implies that EOF has been reached.
+    ///
+    /// If no data is available for reading, the method returns `Poll::Pending`
+    /// and arranges for the current task (via cx.waker()) to receive a notification
+    /// when the stream becomes readable or is closed.
     fn poll_read_chunk(
         &mut self,
         cx: &mut Context,
@@ -187,6 +195,8 @@ impl RecvStream {
     ///
     /// Slightly more efficient than `read` due to not copying. Chunk boundaries
     /// do not correspond to peer writes, and hence cannot be used as framing.
+    ///
+    /// This operation is cancel-safe.
     pub async fn read_chunks(&mut self, bufs: &mut [Bytes]) -> Result<Option<usize>, ReadError> {
         ReadChunks { stream: self, bufs }.await
     }
@@ -229,13 +239,15 @@ impl RecvStream {
     /// If unordered reads have already been made, the resulting buffer may have gaps containing
     /// arbitrary data.
     ///
+    /// This operation is *not* cancel-safe.
+    ///
     /// [`ReadToEndError::TooLong`]: crate::ReadToEndError::TooLong
     pub async fn read_to_end(&mut self, size_limit: usize) -> Result<Vec<u8>, ReadToEndError> {
         ReadToEnd {
             stream: self,
             size_limit,
             read: Vec::new(),
-            start: u64::max_value(),
+            start: u64::MAX,
             end: 0,
         }
         .await
@@ -244,8 +256,8 @@ impl RecvStream {
     /// Stop accepting data
     ///
     /// Discards unread data and notifies the peer to stop transmitting. Once stopped, further
-    /// attempts to operate on a stream will yield `UnknownStream` errors.
-    pub fn stop(&mut self, error_code: VarInt) -> Result<(), UnknownStream> {
+    /// attempts to operate on a stream will yield `ClosedStream` errors.
+    pub fn stop(&mut self, error_code: VarInt) -> Result<(), ClosedStream> {
         let mut conn = self.conn.state.lock("RecvStream::stop");
         if self.is_0rtt && conn.check_0rtt().is_err() {
             return Ok(());
@@ -269,6 +281,49 @@ impl RecvStream {
         self.stream
     }
 
+    /// Completes when the stream has been reset by the peer or otherwise closed
+    ///
+    /// Yields `Some` with the reset error code when the stream is reset by the peer. Yields `None`
+    /// when the stream was previously [`stop()`](Self::stop)ed, or when the stream was
+    /// [`finish()`](crate::SendStream::finish)ed by the peer and all data has been received, after
+    /// which it is no longer meaningful for the stream to be reset.
+    ///
+    /// This operation is cancel-safe.
+    pub async fn received_reset(&mut self) -> Result<Option<VarInt>, ResetError> {
+        poll_fn(|cx| {
+            let mut conn = self.conn.state.lock("RecvStream::reset");
+            if self.is_0rtt && conn.check_0rtt().is_err() {
+                return Poll::Ready(Err(ResetError::ZeroRttRejected));
+            }
+
+            if let Some(code) = self.reset {
+                return Poll::Ready(Ok(Some(code)));
+            }
+
+            match conn.inner.recv_stream(self.stream).received_reset() {
+                Err(_) => Poll::Ready(Ok(None)),
+                Ok(Some(error_code)) => {
+                    // Stream state has just now been freed, so the connection may need to issue new
+                    // stream ID flow control credit
+                    conn.wake();
+                    Poll::Ready(Ok(Some(error_code)))
+                }
+                Ok(None) => {
+                    if let Some(e) = &conn.error {
+                        return Poll::Ready(Err(e.clone().into()));
+                    }
+                    // Resets always notify readers, since a reset is an immediate read error. We
+                    // could introduce a dedicated channel to reduce the risk of spurious wakeups,
+                    // but that increased complexity is probably not justified, as an application
+                    // that is expecting a reset is not likely to receive large amounts of data.
+                    conn.blocked_readers.insert(self.stream, cx.waker().clone());
+                    Poll::Pending
+                }
+            }
+        })
+        .await
+    }
+
     /// Handle common logic related to reading out of a receive stream
     ///
     /// This takes an `FnMut` closure that takes care of the actual reading process, matching
@@ -296,7 +351,7 @@ impl RecvStream {
 
         // If we stored an error during a previous call, return it now. This can happen if a
         // `read_fn` both wants to return data and also returns an error in its final stream status.
-        let status = match self.reset.take() {
+        let status = match self.reset {
             Some(code) => ReadStatus::Failed(None, Reset(code)),
             None => {
                 let mut recv = conn.inner.recv_stream(self.stream);
@@ -328,6 +383,7 @@ impl RecvStream {
             ReadStatus::Failed(read, Reset(error_code)) => match read {
                 None => {
                     self.all_data_read = true;
+                    self.reset = Some(error_code);
                     Poll::Ready(Err(ReadError::Reset(error_code)))
                 }
                 done => {
@@ -417,7 +473,7 @@ impl futures_io::AsyncRead for RecvStream {
         buf: &mut [u8],
     ) -> Poll<io::Result<usize>> {
         let mut buf = ReadBuf::new(buf);
-        ready!(RecvStream::poll_read(self.get_mut(), cx, &mut buf))?;
+        ready!(Self::poll_read_buf(self.get_mut(), cx, &mut buf))?;
         Poll::Ready(Ok(buf.filled().len()))
     }
 }
@@ -428,7 +484,7 @@ impl tokio::io::AsyncRead for RecvStream {
         cx: &mut Context<'_>,
         buf: &mut ReadBuf<'_>,
     ) -> Poll<io::Result<()>> {
-        ready!(Self::poll_read(self.get_mut(), cx, buf))?;
+        ready!(Self::poll_read_buf(self.get_mut(), cx, buf))?;
         Poll::Ready(Ok(()))
     }
 }
@@ -444,7 +500,7 @@ impl Drop for RecvStream {
             return;
         }
         if !self.all_data_read {
-            // Ignore UnknownStream errors
+            // Ignore ClosedStream errors
             let _ = conn.inner.recv_stream(self.stream).stop(0u32.into());
             conn.wake();
         }
@@ -463,8 +519,8 @@ pub enum ReadError {
     #[error("connection lost")]
     ConnectionLost(#[from] ConnectionError),
     /// The stream has already been stopped, finished, or reset
-    #[error("unknown stream")]
-    UnknownStream,
+    #[error("closed stream")]
+    ClosedStream,
     /// Attempted an ordered read following an unordered read
     ///
     /// Performing an unordered read allows discontinuities to arise in the receive buffer of a
@@ -484,24 +540,60 @@ pub enum ReadError {
 impl From<ReadableError> for ReadError {
     fn from(e: ReadableError) -> Self {
         match e {
-            ReadableError::UnknownStream => Self::UnknownStream,
+            ReadableError::ClosedStream => Self::ClosedStream,
             ReadableError::IllegalOrderedRead => Self::IllegalOrderedRead,
         }
     }
 }
 
+impl From<ResetError> for ReadError {
+    fn from(e: ResetError) -> Self {
+        match e {
+            ResetError::ConnectionLost(e) => Self::ConnectionLost(e),
+            ResetError::ZeroRttRejected => Self::ZeroRttRejected,
+        }
+    }
+}
+
 impl From<ReadError> for io::Error {
     fn from(x: ReadError) -> Self {
         use self::ReadError::*;
         let kind = match x {
             Reset { .. } | ZeroRttRejected => io::ErrorKind::ConnectionReset,
-            ConnectionLost(_) | UnknownStream => io::ErrorKind::NotConnected,
+            ConnectionLost(_) | ClosedStream => io::ErrorKind::NotConnected,
             IllegalOrderedRead => io::ErrorKind::InvalidInput,
         };
         Self::new(kind, x)
     }
 }
 
+/// Errors that arise while waiting for a stream to be reset
+#[derive(Debug, Error, Clone, PartialEq, Eq)]
+pub enum ResetError {
+    /// The connection was lost
+    #[error("connection lost")]
+    ConnectionLost(#[from] ConnectionError),
+    /// This was a 0-RTT stream and the server rejected it
+    ///
+    /// Can only occur on clients for 0-RTT streams, which can be opened using
+    /// [`Connecting::into_0rtt()`].
+    ///
+    /// [`Connecting::into_0rtt()`]: crate::Connecting::into_0rtt()
+    #[error("0-RTT rejected")]
+    ZeroRttRejected,
+}
+
+impl From<ResetError> for io::Error {
+    fn from(x: ResetError) -> Self {
+        use ResetError::*;
+        let kind = match x {
+            ZeroRttRejected => io::ErrorKind::ConnectionReset,
+            ConnectionLost(_) => io::ErrorKind::NotConnected,
+        };
+        Self::new(kind, x)
+    }
+}
+
 /// Future produced by [`RecvStream::read()`].
 ///
 /// [`RecvStream::read()`]: crate::RecvStream::read
@@ -511,12 +603,12 @@ struct Read<'a> {
     buf: ReadBuf<'a>,
 }
 
-impl<'a> Future for Read<'a> {
+impl Future for Read<'_> {
     type Output = Result<Option<usize>, ReadError>;
 
     fn poll(self: Pin<&mut Self>, cx: &mut Context) -> Poll<Self::Output> {
         let this = self.get_mut();
-        ready!(this.stream.poll_read(cx, &mut this.buf))?;
+        ready!(this.stream.poll_read_buf(cx, &mut this.buf))?;
         match this.buf.filled().len() {
             0 if this.buf.capacity() != 0 => Poll::Ready(Ok(None)),
             n => Poll::Ready(Ok(Some(n))),
@@ -533,16 +625,16 @@ struct ReadExact<'a> {
     buf: ReadBuf<'a>,
 }
 
-impl<'a> Future for ReadExact<'a> {
+impl Future for ReadExact<'_> {
     type Output = Result<(), ReadExactError>;
     fn poll(self: Pin<&mut Self>, cx: &mut Context) -> Poll<Self::Output> {
         let this = self.get_mut();
         let mut remaining = this.buf.remaining();
         while remaining > 0 {
-            ready!(this.stream.poll_read(cx, &mut this.buf))?;
+            ready!(this.stream.poll_read_buf(cx, &mut this.buf))?;
             let new = this.buf.remaining();
             if new == remaining {
-                return Poll::Ready(Err(ReadExactError::FinishedEarly));
+                return Poll::Ready(Err(ReadExactError::FinishedEarly(this.buf.filled().len())));
             }
             remaining = new;
         }
@@ -554,8 +646,8 @@ impl<'a> Future for ReadExact<'a> {
 #[derive(Debug, Error, Clone, PartialEq, Eq)]
 pub enum ReadExactError {
     /// The stream finished before all bytes were read
-    #[error("stream finished early")]
-    FinishedEarly,
+    #[error("stream finished early ({0} bytes read)")]
+    FinishedEarly(usize),
     /// A read error occurred
     #[error(transparent)]
     ReadError(#[from] ReadError),
@@ -571,7 +663,7 @@ struct ReadChunk<'a> {
     ordered: bool,
 }
 
-impl<'a> Future for ReadChunk<'a> {
+impl Future for ReadChunk<'_> {
     type Output = Result<Option<Chunk>, ReadError>;
     fn poll(mut self: Pin<&mut Self>, cx: &mut Context) -> Poll<Self::Output> {
         let (max_length, ordered) = (self.max_length, self.ordered);
@@ -588,7 +680,7 @@ struct ReadChunks<'a> {
     bufs: &'a mut [Bytes],
 }
 
-impl<'a> Future for ReadChunks<'a> {
+impl Future for ReadChunks<'_> {
     type Output = Result<Option<usize>, ReadError>;
     fn poll(self: Pin<&mut Self>, cx: &mut Context) -> Poll<Self::Output> {
         let this = self.get_mut();
