@@ -2,6 +2,7 @@ use std::{
     future::Future,
     io,
     pin::Pin,
+    sync::Arc,
     task::{Context, Poll},
     time::Instant,
 };
@@ -11,7 +12,7 @@ use tokio::{
     time::{sleep_until, Sleep},
 };
 
-use super::{AsyncTimer, AsyncUdpSocket, Runtime};
+use super::{AsyncTimer, AsyncUdpSocket, Runtime, UdpPollHelper};
 
 /// A Quinn runtime for Tokio
 #[derive(Debug)]
@@ -26,13 +27,16 @@ impl Runtime for TokioRuntime {
         tokio::spawn(future);
     }
 
-    fn wrap_udp_socket(&self, sock: std::net::UdpSocket) -> io::Result<Box<dyn AsyncUdpSocket>> {
-        udp::UdpSocketState::configure((&sock).into())?;
-        Ok(Box::new(UdpSocket {
+    fn wrap_udp_socket(&self, sock: std::net::UdpSocket) -> io::Result<Arc<dyn AsyncUdpSocket>> {
+        Ok(Arc::new(UdpSocket {
+            inner: udp::UdpSocketState::new((&sock).into())?,
             io: tokio::net::UdpSocket::from_std(sock)?,
-            inner: udp::UdpSocketState::new(),
         }))
     }
+
+    fn now(&self) -> Instant {
+        tokio::time::Instant::now().into_std()
+    }
 }
 
 impl AsyncTimer for Sleep {
@@ -51,22 +55,17 @@ struct UdpSocket {
 }
 
 impl AsyncUdpSocket for UdpSocket {
-    fn poll_send(
-        &self,
-        state: &udp::UdpState,
-        cx: &mut Context,
-        transmits: &[udp::Transmit],
-    ) -> Poll<io::Result<usize>> {
-        let inner = &self.inner;
-        let io = &self.io;
-        loop {
-            ready!(io.poll_send_ready(cx))?;
-            if let Ok(res) = io.try_io(Interest::WRITABLE, || {
-                inner.send(io.into(), state, transmits)
-            }) {
-                return Poll::Ready(Ok(res));
-            }
-        }
+    fn create_io_poller(self: Arc<Self>) -> Pin<Box<dyn super::UdpPoller>> {
+        Box::pin(UdpPollHelper::new(move || {
+            let socket = self.clone();
+            async move { socket.io.writable().await }
+        }))
+    }
+
+    fn try_send(&self, transmit: &udp::Transmit) -> io::Result<()> {
+        self.io.try_io(Interest::WRITABLE, || {
+            self.inner.send((&self.io).into(), transmit)
+        })
     }
 
     fn poll_recv(
@@ -90,6 +89,14 @@ impl AsyncUdpSocket for UdpSocket {
     }
 
     fn may_fragment(&self) -> bool {
-        udp::may_fragment()
+        self.inner.may_fragment()
+    }
+
+    fn max_transmit_segments(&self) -> usize {
+        self.inner.max_gso_segments()
+    }
+
+    fn max_receive_segments(&self) -> usize {
+        self.inner.gro_segments()
     }
 }
